<?php
/* 
 * Copyright (C) 2008 Miroslav Jezbera <jezz@hkfree.org>
 * This file is part of LQueues - simple bandwidth manager for linux.
 *
 * LQueues is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * LQueues is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with LQueues.  If not, see <http://www.gnu.org/licenses/agpl.html>.
 */

$configFile = 'queues.conf';
$statsUrl = '/queues/queues-stats.pl';
$updateCmd = '/var/www/queues/queues-update.sh';
$reloadCmd = '/var/www/queues/queues-restart.sh';

$allowNonAdmin = true;
$adminPassword = '$6$D9UnSguO$DNU70U3U23PgQSbaSHNj8naBWYh.skOqNyYcnnVglJtc2LTa0tlVxX7bNDnrMkx9OssamnmolE7b81.ltEj1M.';
$maxSpeed = 90112;
$historyLength = 10;

$isUpdate = ($_POST and $_POST['id']);
$isDelete = ($_POST and $_POST['delete_id']);
$updateSuccessfull = true;
$updateErrors = array();
if ($isUpdate or $isDelete or !$allowNonAdmin)
	{
	if (!isset($_SERVER['PHP_AUTH_USER']) or $_SERVER['PHP_AUTH_USER'] != 'admin' or crypt($_SERVER['PHP_AUTH_PW'], $adminPassword) != $adminPassword)
		{
		header('WWW-Authenticate: Basic realm="Administrace front"');
		header('HTTP/1.0 401 Unauthorized');
		echo '<html><head><meta http-equiv="content-type" content="text/html; charset=UTF-8"/><title>Přístup zamítnut!</title></head><body><h1>Přístup zamítnut!</h1></body></html>';
		exit;
		}
	}

function addUpdateError($error)
	{
	global $updateSuccessfull, $updateErrors;
	$updateSuccessfull = false;
	array_push($updateErrors, utf8_encode($error));
	}

function doUpdate()
	{
	global $updateCmd, $reloadCmd, $updateErrors, $updateSuccessfull, $maxSpeed;

	header('Content-Type: application/json');

	# Sanitize input
	# Ints
	$id = $_POST['id'];
	$downRate = $_POST['download'];
	$upRate = $_POST['upload'];
	$crc = $_POST['crc'];
	settype($id, "integer");
	settype($downRate, "integer");
	settype($upRate, "integer");
	settype($crc, "integer");

	# Strings
	$inIface = trim($_POST['in_iface']);
	$outIface = trim($_POST['out_iface']);
	$name = trim(preg_replace('/[,;" ]+/', ' ', $_POST['name']));
	$oldName = preg_replace('/[,;"]/', '', $_POST['old_name']);
	$ips = trim(preg_replace('/[^0-9a-f\.:\/]+/i', ' ', $_POST['ips']));

	$result = array('result' => 'ok', 'messages' => array());

	if (!preg_match('/^(\w+\d+)?;(\w+\d+)?$/', "$inIface;$outIface"))
		{
		addUpdateError("Neplatné rozhraní '$inIface' nebo '$outIface'!");
		}
	if (!preg_match('/^(?:(?:
			(?:[1-9]\d?|1\d{2}|2[0-4]\d|25[0-5])(?:\.(?:\d|[1-9]\d|1\d{2}|2[0-4]\d|25[0-5])){3} # IPv4
			| (?:[0-9a-f]{1,4})(?:(?::[0-9a-f]{1,4}){7}|:(?:[0-9a-f]{1,4}:){0,5}(?::[0-9a-f]{1,4})) # IPv6 - simplified (:: accepted at the end)
			)(?:\/\d{2})? # subnet XX
			\s)+$/xi', "$ips "))
		{
		addUpdateError("Chyba ve formátu IP adres: $ips!");
		}

	if ($updateSuccessfull)
		{
		updateConfig($id, $name, $oldName, $inIface, $outIface, $downRate, $upRate, $ips);
		}
	if ($updateSuccessfull)
		{
		# CRC is not for security, but to detect big changes
		if ($id > 0 and crc32("$inIface,$outIface,$ips") == $crc)
			{
			if ($downRate == 0) $downRate = $maxSpeed;
			if ($upRate == 0) $upRate = $maxSpeed;
			if ($inIface)
				{
				$lastline = system("$updateCmd $id $inIface $downRate 2>&1", $retval);
				}
			if ($outIface)
				{
				$lastline = system("$updateCmd $id $outIface $upRate 2>&1", $retval);
				}
			if ($retval === FALSE || $retval != 0)
				{
				addUpdateError("Chyba při nastavení limitu! Code: $retval ($lastline)");
				}
			else
				echo '{ "result": "ok" }';
			}
		else
			{
			system("$reloadCmd");
			echo '{ "result": "redirect" }';
			}
		}
	if (!$updateSuccessfull)
		echo '{ "result": "fail", "errors": ' . json_encode($updateErrors) . " }";
	exit;
	}

function updateConfig($id, $name, $oldName, $inIface, $outIface, $downRate, $upRate, $ips)
	{
	global $configFile, $maxSpeed;

	$semid = sem_get(ftok($configFile, 'q'), 1, 0600);
	if ($semid < 0)
		{
		addUpdateError("Can't open semaphore!");
		return false;
		}
	if (!sem_acquire($semid))
		{
		addUpdateError("Can't acquire semaphore!");
		return false;
		}
	if (!copy($configFile, "$configFile.bak"))
		{
		sem_release($semid);
		addUpdateError("Can't backup config file!");
		return false;
		}
	$config = fopen($configFile, 'r+');
	if (!$config)
		{
		sem_release($semid);
		addUpdateError("Can't open config file for writting!");
		return false;
		}
	$lineNum = 0;
	$pos = 0;
	$newline = '';
	# no IP = delete entry
	if (strlen($ips) > 0)
		{
		$newline = "$name,$inIface,$outIface,";
		if ($downRate > 0 and $downRate < $maxSpeed) $newline .= $downRate;
		$newline .= ',';
		if ($upRate > 0 and $upRate < $maxSpeed) $newline .= $upRate;
		$newline .= ",$ips\n";
		}
	if ($id < 0)
		{
		# add new queue
		fseek($config, 0, SEEK_END);
		fwrite($config, "$newline");
		}
	else
		{
		$line = '';
		while (!feof($config) and $lineNum < $id)
			{
			$pos = ftell($config);
			$line = fgets($config);
			if (preg_match('/^\s*([%#].*)?$/', $line)) continue;
			$lineNum++;
			}
		fseek($config, $pos, SEEK_SET);
		if (feof($config) or $lineNum < $id or trim(substr($line,0, strpos($line, ','))) != $oldName)
			{
			sem_release($semid);
			fclose($config);
			addUpdateError('Konfigurační soubor byl změněn (id nebylo nalezeno nebo se neshoduje jméno)!');
			return false;
			}
		$orig = fopen("$configFile.bak", 'r');
		if (!$orig)
			{
			sem_release($semid);
			fclose($config);
			addUpdateError("Can't open backup file!");
			return false;
			}
		fseek($orig, $pos, SEEK_SET);
		fgets($orig); # move after line
		fwrite($config, $newline);
		ftruncate($config, ftell($config));
		while (!feof($orig))
			{
			$line = trim(fgets($orig));
			if ($line != '')
				{
				fwrite($config, "$line\n");
				}
			}
		fclose($orig);
		}
	sem_release($semid);
	fclose($config);
	return true;
	}

if ($isUpdate)
	{
	doUpdate();
	exit;
	}
?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
	<meta http-equiv="content-type" content="text/html; charset=utf-8" />
	<title>Queues</title>
	<link rel="StyleSheet" href="css/main.css" type="text/css" />
	<link rel="StyleSheet" href="css/jquery-ui-1.8.20.custom.css" type="text/css" />
	<script type="text/javascript" src="js/jquery-1.7.1.min.js"></script>
	<script type="text/javascript" src="js/jquery-ui-1.8.20.custom.min.js"></script>
	<script type="text/javascript" src="js/ui.spinner.min.js"></script>
	<script type="text/javascript" src="js/raphael-min.js"></script>
	<script type="text/javascript">
// Global variables
maxSpeed = <?php echo $maxSpeed; ?>;
speedHistoryLength = <?php echo $historyLength; ?>;
users = {};
	</script>
	<script type="text/javascript" src="js/scripts.js"></script>
	<script type="text/javascript">
//<![CDATA[
<?php
$devices = array();
$availDevices = array();

function processConfigFile()
	{
	global $configFile;

	$config = fopen($configFile, 'r');
	if (!$config) die(']]></script></head><body><p class="error">Can\'t open config file!</p></body></html>');

	$lineNum = 1;
	while (!feof($config))
		{
		$line = fgets($config);
		if (preg_match('/^\s*(#.*)?$/', $line)) continue;
		if (preg_match('/^\s*%\s*(\b[^=]+\b)\s*=\s*(\b.+\b)\s*$/', $line, $matches))
			{
			processDirective($matches[1], $matches[2]);
			continue;
			}
		$parts = preg_split('/\s*,\s*/', $line);
		addUser($lineNum, ltrim($parts[0]), $parts[1], $parts[2], rtrim($parts[5]));
		$lineNum++;
		}
	listDevices();

	fclose($config);
	}

function addUser($id,$userName,$inIface,$outIface,$ips)
	{
	global $devices;
	if ($inIface != '')
		{
		$devices[$inIface] = 1;
		}
	if ($outIface != '')
		{
		$devices[$outIface] = 1;
		}
	echo "users[$id] = new User($id, \"$userName\",'$inIface','$outIface','$ips'," . crc32("$inIface,$outIface,$ips") . ");\n";
	}

function listDevices()
	{
	global $devices;
	echo 'devices = new Array(' . implode(", ", array_map("quote" ,array_keys($devices))) . ");\n";
	}

function quote($str)
	{
	return "'$str'";
	}

function processDirective($var,$value)
	{
	global $availDevices;
	if ($var == 'devices')
		{
		$parts = preg_split('/\s*,\s*/', $value, -1, PREG_SPLIT_NO_EMPTY);
		foreach ($parts as $value)
			{
			$availDevices[$value] = 1;
			}
		}
	}

function fatalErrorInScript($str)
	{
	echo "//]]></script></head><body>";
	die("<p class=\"error\">$str</p></body>");
	}

# main
echo "statsUrl = '$statsUrl';\n";

if ($isDelete)
	{
	# Sanitize input
	# Ints
	$id = $_POST['delete_id'];
	settype($id, "integer");

	# Strings
	$oldName = preg_replace('/[,;"]/', '', $_POST['old_name']);

	updateConfig($id, '', $oldName, '', '', 0, 0, '');
	if ($updateSuccessfull)
		{
		system("$reloadCmd");
		}
	}

processConfigFile();
?>

$(function ()
	{
	fillQueuesList();
	loadSettings();
	initFilters();
	initLimitsDialog();
	initSettingsPanel();
	setTimer();
	});
//]]>
	</script>
	<style type="text/css">
.filtered .item { display: none; }
<?php
function createSelectorForDevice($device, $element)
	{
	return ".filtered.$device $element.item.$device";
	}

function createRowStyles($element, $style)
	{
	global $devices;
	echo implode(
		",\n",
		array_map(
			function ($device) use ($element) { return createSelectorForDevice($device, $element); },
			array_keys($devices))
		) . " { $style }\n";
	}

createRowStyles('tr', 'display: table-row;');
createRowStyles('div', 'display: block;');
?>
	</style>
</head>
<body>
	<h1>Fronty</h1>

	<!-- Settings -->
	<div id="settings" class="ui-widget ui-widget-content ui-corner-all">
		<h3 class="ui-widget-header ui-corner-all">Nastavení</h3>
		<div>
			<h4>Aktualizace</h4>
			<label for="interval">Interval:</label>
			<input type="text" id="interval" class="num" size="3" value="10" />
			<button id="changeIntervalButton">Změnit</button>
			<button id="pauseButton">Pozastavit</button>
		</div>
		<div>
			<h4>Grafy</h4>
			<label for="history-length">Historie:</label>
			<input type="text" id="history-length" class="num" size="3" value="<?php echo $historyLength; ?>" />
			<button id="change-history-length">Změnit</button>
		</div>
		<div id="filters">
			<h4>Filtry</h4>
		</div>
		<button id="addQueueButton">Přidat novou frontu</button>
	</div>

	<!-- Graphs -->
	<div id="graphs" class="filtered ui-corner-all">
	</div>

	<!-- Queues list -->
	<table id="queues-list" class="filtered ui-widget ui-widget-content ui-corner-all">
		<thead>
			<tr class="header ui-widget-header ui-corner-top"><th class="ui-corner-tl">Jméno</th><th>IP</th><th>Max. upload</th><th>Max. download</th><th>Upload</th><th>Download</th><th class="ui-corner-tr"><a href="javascript:void(0)" class="add-queue" title="Přidat frontu"><span class="ui-icon ui-icon-plus">Přidat</span></a></th></tr>
		</thead>
		<tbody>
		</tbody>
	</table>

	<!-- Limits dialog -->
	<div id="limits-dialog" style="display: none">
		<form id="change-form" method="post">
			<input type="hidden" id="id" name="id" />
			<input type="hidden" id="old_name" name="old_name" />
			<input type="hidden" id="crc" name="crc" />
			<div id="user">
				<div>
					<label for="name">Jméno:</label> <input type="text" name="name" id="name" />
				</div>
				<div>
					<label for="ips">IP:</label> <input type="text" id="ips" name="ips" />
				</div>
			</div>

			<div id="rates">
				<div>
					<div class="device">
						<label for="out_iface">Out device:</label>
						<select name="out_iface" id="out_iface"></select>
					</div>
					<div class="limit">
						<label for="upload">Upload: </label>
						<div><input type="text" size="4" id="upload" name="upload" value="0" /> kbit/s</div>
					</div>
				</div>
				<div>
					<div class="device">
						<label for="in_iface">In device:</label>
						<select name="in_iface" id="in_iface"></select>
					</div>
					<div class="limit">
						<label for="download">Download: </label>
						<div><input type="text" size="4" id="download" name="download" value="0" /> kbit/s</div>
					</div>
				</div>
			</div>

		</form>
		<form id="delete-form" method="post">
			<input type="hidden" id="delete_id" name="delete_id" />
			<input type="hidden" id="delete_old_name" name="old_name" />
		</form>
	</div>

	<div id="dialog-confirm" title="Smazat frontu?" style="display: none">
		<p><span class="ui-icon ui-icon-alert"></span><span class="question"></span></p>
	</div>
</body>
</html>
