/* 
 * Copyright (C) 2008 Miroslav Jezbera <jezz@hkfree.org>
 * This file is part of LQueues - simple bandwidth manager for linux.
 *
 * LQueues is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * LQueues is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with LQueues.  If not, see <http://www.gnu.org/licenses/agpl.html>.
 */

var nextCounter = 0;

/*
 * Queue object
 */
function Queue()
	{
	this.lastCounter = -1;
	this.lastUpdateTime = 0;
	this.bytes = 0;
	this.speeds = reallocateArray(0, speedHistoryLength);
	this.rate = 0;
	}

Queue.prototype.assignInfoElements = function ($speedInfo, $rateInfo)
	{
	this.$speedInfo = $speedInfo;
	this.$rateInfo = $rateInfo;
	};

Queue.prototype.update = function (counter, bytes, rate, time)
	{
	if (counter <= this.lastCounter)
		return;
	var speed = (bytes - this.bytes) * 8 / ((time - this.lastUpdateTime) * 1024);
	this.lastUpdateTime = time;
	this.updateSpeed(counter, bytes, speed);
	this.updateRate(rate);
	};

Queue.prototype.updateSpeed = function (counter, bytes, speed)
	{
	if (bytes != this.bytes || this.speeds[this.speeds.length - 1] != 0)
		{
		this.bytes = bytes;
		this.$speedInfo.text(speed.toFixed(1) + ' kbit/s');
		}
	this.addMissingSpeedRecords(counter);
	this.speeds.shift();
	this.speeds.push(speed);
	};

Queue.prototype.updateRate = function (rate)
	{
	if (rate != this.rate)
		{
		this.rate = rate;
		this.$rateInfo.text(formatRate(rate, 'kbit/s'));
		}
	};

Queue.prototype.addMissingSpeedRecords = function (counter)
	{
	if (counter <= this.lastCounter)
		return;
	while (++this.lastCounter < counter)
		{
		this.speeds.shift();
		this.speeds.push(0);
		}
	};

Queue.prototype.changeHistoryLength = function (length)
	{
	if (length < 2)
		return;
	this.speeds = reallocateArray(0, length, this.speeds);
	};

/*
 * User object
 */
function User(id, name,inIface,outIface,ips,crc)
	{
	this.id = id;
	this.name = name;
	this.ips = ips;
	this.crc = crc;
	this.inIface = inIface;
	this.outIface = outIface;
	this.uploadQueue = new Queue();
	this.downloadQueue = new Queue();

	this.updateCallback = null;
	}

User.prototype.createRow = function ()
	{
	var self = this;
	var $name_th = $('<th><a href="javascript:void(0)" class="name">' + this.name + '</a></th>');
	var $ips = $('<td class="ips"><div title="' + this.ips + '">' + this.ips + '</div></div></td>');
	var $up_rate = $('<td class="up_rate">unlimited</td>');
	var $up_speed = $('<td class="up_speed">0.0 kbit/s</td>');
	var $down_rate = $('<td class="down_rate">unlimited</td>');
	var $down_speed = $('<td class="down_speed">0.0 kbit/s</td>');
	var $actions = $(
		'<td class="actions">' +
			'<a href="javascript:void(0)" class="delete-queue"><span class="ui-icon ui-icon-trash">smazat</span></a>' +
			'<a href="javascript:void(0)" class="show-graph"><span class="ui-icon">graf</span></a>' +
		'</td>');
	var $row = $('<tr class="item"></tr>');
	if (!this.inIface)
		{
		$down_rate.addClass('disabled');
		$down_speed.addClass('disabled');
		}
	else
		{
		$row.addClass(this.inIface);
		}
	if (!this.outIface)
		{
		$up_rate.addClass('disabled');
		$up_speed.addClass('disabled');
		}
	else
		{
		$row.addClass(this.outIface);
		}

	$row
		.append($name_th)
		.append($ips)
		.append($up_rate)
		.append($down_rate)
		.append($up_speed)
		.append($down_speed)
		.append($actions);

	$row.find('a.name').click(function(ev) { showLimits(self.id); });
	$row.find('a.show-graph').click(function() { createGraph(self.id); });
	$row.find('a.delete-queue').click(function() { deleteQueue(self.id); });

	this.$row = $row;
	this.uploadQueue.assignInfoElements($up_speed, $up_rate);
	this.downloadQueue.assignInfoElements($down_speed, $down_rate);

	return $row;
	};

User.prototype.onUpdate = function (callback)
	{
	this.updateCallback = callback;
	};

User.prototype.update = function (iface, counter, bytes, rate, time)
	{
	if (iface == this.inIface)
		this.downloadQueue.update(counter, bytes, rate, time);
	else if (iface == this.outIface)
		this.uploadQueue.update(counter, bytes, rate, time);
	else
		{
		console.log('Unknown iface: ' + iface);
		return;
		}
	if (this.updateCallback && this.uploadQueue.counter == this.downloadQueue.counter)
		this.updateCallback(this);
	};

User.prototype.getInterfaces = function ()
	{
	var ifaces = [];
	if (this.inIface)
		ifaces.push(this.inIface);
	if (this.outIface)
		ifaces.push(this.outIface);
	return ifaces;
	};

User.prototype.changeHistoryLength = function (length)
	{
	this.downloadQueue.changeHistoryLength(length);
	this.uploadQueue.changeHistoryLength(length);
	};

function reallocateArray(value, count, oldArray)
	{
	var array = new Array();
	if (oldArray)
		{
		while (count-- > 0 && oldArray.length > 0)
			array.unshift(oldArray.pop());
		}
	while (count-- > 0)
		array.unshift(value);
	return array;
	}

function updateStatistics()
	{
	for (var device in activeDevices)
		{
		$.get(statsUrl, { device: device, counter: nextCounter }, onXmlLoaded);
		}
	nextCounter++;
	}

function onXmlLoaded(data)
	{
	var $queues = $(data).find('queues');
	if ($queues.length == 0) return;
	var time = $queues.find('time').text();
	var device = $queues.find('device').text();
	var counter = parseInt($queues.find('counter').text());
	var $classes = $queues.find('class');

	$classes.each(function ()
		{
		var $Class = $(this);
		var id = $Class.attr('id');
		if (id != '0' && id != 'ffff' && id != 'fffe')
			{
			var rate = parseInt($Class.find('rate').text());
			var bytes = parseInt($Class.find('bytes').text());

			users[id].update(device, counter, bytes, rate, time);
			}
		});
	}

var updateTimer = 0;
var activeDevices = {};
var activeFilters = [];
var linkedDevices = {};
var settings;

function loadSettings()
	{
	settings = parseHash();
	if (settings['filters'])
		activeFilters = settings['filters'].split(',');
	}

function initSettingsPanel()
	{
	$('#settings button').button();
	$('#changeIntervalButton').click(setTimer);
	$('#pauseButton').click(pause);
	$('#addQueueButton, .add-queue').click(function (event) { showLimits(-1); });
	$('#interval').spinner({ min: 1 });
	$('#history-length').spinner({ min: 5 });
	$('#change-history-length').click(changeHistoryLength);
	}

function setTimer()
	{
	updateStatistics();
	if (updateTimer) clearInterval(updateTimer);
	updateTimer = window.setInterval(updateStatistics, $('#interval').val() * 1000);
	}

function changeHistoryLength()
	{
	$.each(users, function(idx, user)
		{
		user.changeHistoryLength($('#history-length').val());
		});
	}

function pause()
	{
	var $buttonText = $('#pauseButton .ui-button-text');
	if (updateTimer)
		{
		clearInterval(updateTimer);
		updateTimer = null;
		$buttonText.text('Spustit');
		}
	else
		{
		setTimer();
		$buttonText.text('Pozastavit');
		}
	}

function fillQueuesList()
	{
	var $queues_list = $('#queues-list tbody');
	$.each(users, function(idx, user)
		{
		$queues_list.append(user.createRow());
		});
	}

function initFilters()
	{
	var template = '<label><input id="#" type="checkbox" /><span>#</span></label>';
	var $filters = $('#filters');

	for (var idx = 0; idx < devices.length; ++idx)
		{
		linkedDevices[devices[idx]] = {};
		$filters.append(template.replace(/#/g, devices[idx]));
		}
	if (activeFilters.length == 0)
		activeFilters = devices;
	$('#' + activeFilters.join(',#')).attr('checked', 'checked');
	$filters.find(':checkbox').change(function ()
		{
		var id = $(this).attr('id');
		$('.filtered').toggleClass(id);
		refreshActiveDevices();
		});
	for (var id in users)
		{
		if (!users[id].inIface || !users[id].outIface)
			continue;
		linkedDevices[users[id].inIface][users[id].outIface] = true;
		linkedDevices[users[id].outIface][users[id].inIface] = true;
		}
	$('.filtered').addClass(activeFilters.join(' '));
	refreshActiveDevices();
	}

function refreshActiveDevices()
	{
	var newActiveDevices = {};
	activeFilters = [];
	$('#filters :checked').each(function ()
		{
		var device = $(this).attr('id');
		activeFilters.push(device);
		newActiveDevices[device] = true;
		$.extend(newActiveDevices, linkedDevices[device]);
		});
	activeDevices = newActiveDevices;
	updateHash('filters', activeFilters.length < devices.length ? activeFilters.join() : null);
	}

function keys(object)
	{
	var result = [];
	for (var key in object)
		result.push(key);
	return result;
	}

function updateHash(property, value)
	{
	var regexp = new RegExp('^#(.*;){0,1}' + property + '=[^;]*(;.*){0,1}$');
	var emptyValue = value === undefined || value === null || String(value).length == 0;
	var hashProperty = emptyValue ? '' : property + '=' + value;
	if (regexp.test(window.location.hash))
		{
		window.location.hash = window.location.hash.replace(regexp, '$1' + hashProperty + '$2');
		}
	else if (!emptyValue)
		{
		if (window.location.hash)
			window.location.hash += ';' + hashProperty;
		else
			window.location.hash = hashProperty;
		}
	}

function parseHash()
	{
	var hash = window.location.hash.substr(1);
	var parts = hash.split(';');
	var result = {};
	for (var idx = 0; idx < parts.length; idx++)
		{
		var equalPos = parts[idx].indexOf('=');
		var key = parts[idx].substr(0, equalPos);
		var value = parts[idx].substr(equalPos + 1);
		result[key] = value;
		}
	return result;
	}

function initLimitsDialog()
	{
	$('#limits-dialog').dialog(
		{
		autoOpen: false,
		modal: true,
		width: '32em',
		buttons:
			[
				{
				text: 'Změnit',
				click: function () { submitQueueChanges(); }
				},
				{
				text: 'Zavřít',
				click: closeLimits
				}
			]
		});
	var $in_iface = $('#in_iface'),
		$out_iface = $('#out_iface');
	for (var idx = 0; idx < devices.length; ++idx)
		{
		var option = '<option value="' + devices[idx] + '">' + devices[idx] + '</option>';
		$in_iface.append(option);
		$out_iface.append(option);
		}
	}

function submitQueueChanges()
	{
	$.ajax({
		url: '',
		type: 'POST',
		data: $('#limits-dialog #change-form').serialize(),
		dataType: 'json',
		success: function (data) {
			if (data.result == 'ok') {
				// will be updated on next refresh
				closeLimits();
			} else if (data.result == 'redirect') {
				window.location.reload();
			} else if (data.result == 'fail') {
				showError(data.message);
			}
		}
	});
	}

function showError(message)
	{
	// TODO
	}

function showLimits(id)
	{
	var $dialog = $('#limits-dialog');
	$('#id').val(id);
	if (id > 0)
		{
		$dialog.dialog('option', 'title', users[id].name);
		$('#name, #old_name').val(users[id].name);
		$('#upload').val(formatRate(users[id].uploadQueue.rate));
		$('#download').val(formatRate(users[id].downloadQueue.rate));
		$('#in_iface').val(users[id].inIface);
		$('#out_iface').val(users[id].outIface);
		$('#ips').val(users[id].ips);
		$('#crc').val(users[id].crc);
		}
	else
		{
		$dialog.dialog('option', 'title', 'Nová fronta');
		$('#name').val('Jméno');
		$('#old_name').val('');
		$('#upload, #download, #crc').val(0);
		$('#in_iface, #out_iface, #ips').val('');
		}
	$dialog.dialog('open');
	}

function formatRate(rate, units)
	{
	if (rate >= maxSpeed)
		return 'unlimited';
	return units === undefined ? rate : rate + ' ' + units;
	}

function closeLimits()
	{
	$('#limits-dialog').dialog('close');
	}

function deleteQueue(id)
	{
	var name = users[id].name;
	$('#dialog-confirm .question').text('Smazat frontu uživatele ' + name + '?');
	$('#delete_old_name').val(name);
	$('#delete_id').val(id);
	$('#dialog-confirm')
		.dialog(
			{
			resizable: false,
			height: 170,
			width: 360,
			modal: true,
			buttons:
				{
					'Ano': function() {
						$(this).dialog('close');
						$('#delete-form').submit();
					},
					'Ne': function() {
						$(this).dialog('close');
					}
				}
			});
	}

function createGraph(userId)
	{
	var userGraphClass = 'user' + userId;
	if ($('#graphs .graph.' + userGraphClass).length)
		return;
	var user = users[userId];
	var $graph = $('<div class="graph item ' + userGraphClass + ' ' + user.getInterfaces().join(' ') + '"></div>');
	var graphObj = new Graph($graph[0], user);
	var graphHeight;
	var graphWidth;
	$graph.prepend('<h4 class="label">' + user.name + '<a href="javascript:void(0)" class="close cross">x</a></h4>');
	$('#graphs').append($graph);
	graphObj.register();
	$graph.resizable({
		minWidth: 410,
		minHeight: 230, // TODO: Synchronize with settings in Graph object
		maxWidth: $('#graphs').innerWidth() - 6, // border + padding
		start: function () {
			graphHeight = $graph.innerHeight();
			graphWidth = $graph.innerWidth();
		},
		stop: function () {
			graphObj.resize($graph.innerWidth() - graphWidth, $graph.innerHeight() - graphHeight);
		}
	});
	$graph.find('a.close').click(function ()
		{
		graphObj.unregister();
		$graph.remove();
		});
	$graph.data('graph', graphObj);
	}

/* Taken from http://raphaeljs.com/analytics.html */
function getAnchors(p1x, p1y, p2x, p2y, p3x, p3y) {
    var l1 = (p2x - p1x) / 2,
        l2 = (p3x - p2x) / 2,
        a = Math.atan((p2x - p1x) / Math.abs(p2y - p1y)),
        b = Math.atan((p3x - p2x) / Math.abs(p2y - p3y));
    a = p1y < p2y ? Math.PI - a : a;
    b = p3y < p2y ? Math.PI - b : b;
    var alpha = Math.PI / 2 - ((a + b) % (Math.PI * 2)) / 2,
        dx1 = l1 * Math.sin(alpha + a),
        dy1 = l1 * Math.cos(alpha + a),
        dx2 = l2 * Math.sin(alpha + b),
        dy2 = l2 * Math.cos(alpha + b);
    return {
        x1: p2x - dx1,
        y1: p2y + dy1,
        x2: p2x + dx2,
        y2: p2y + dy2
    };
}

/*
 * Graph object
 */
function Graph(parent, user)
	{
	this.user = user;
	this.minUnit = 16; // kB/s
	this.parent = parent;
	this.width = 400;
	this.height = 200;
	this.canvas = Raphael(parent, this.width + 1, this.height + 1);
	this.leftMargin = 40;
	this.topMargin = 5;
	this.yAxisMaximum = 0;
	this.parentMargins = $(parent).parent().innerHeight();

	this.draw();
	}

Graph.prototype.register = function ()
	{
	var self = this;
	this.user.onUpdate(function () { self.draw(); });
	};

Graph.prototype.unregister = function ()
	{
	this.user.onUpdate(null);
	};

Graph.prototype.resize = function (deltaX, deltaY)
	{
	var $parent = $(this.parent);
	this.width += deltaX;
	this.height += deltaY;
	this.canvas.setSize(this.width + 1, this.height + 1);
	this.yAxisMaximum = 0;
	this.draw();
	};

Graph.prototype.draw = function ()
	{
	var newMaxYAxis = this.getYAxisMaximum();
	if (this.yAxisMaximum < newMaxYAxis || this.yAxisMaximum / 2 > newMaxYAxis)
		{
		this.canvas.clear();
		this.yAxisMaximum = newMaxYAxis;
		this.drawGrid();
		}
	else if (this.downloadPoints)
		{
		this.downloadPoints.remove();
		this.uploadPoints.remove();
		}
	this.downloadPoints = this.drawPoints(this.user.downloadQueue.speeds, '#5d5');
	this.uploadPoints = this.drawPoints(this.user.uploadQueue.speeds, '#d55');
	};

Graph.prototype.getYAxisMaximum = function ()
	{
	var maxDownload = Math.max.apply(Math, this.user.downloadQueue.speeds);
	var maxUpload = Math.max.apply(Math, this.user.uploadQueue.speeds);
	return Math.ceil(Math.max.apply(Math, [ maxDownload * 1.2, maxUpload * 1.2, this.minUnit ]) / this.minUnit) * this.minUnit;
	};

Graph.prototype.drawGrid = function ()
	{
	var path = [ 'M', this.leftMargin + 0.5, 0.5, 'V', this.height + 0.5, 'H', this.width + 0.5 ];
	this.canvas.path(path.join()).attr({ stroke: '#aaa' });
	var labelDensity = this.computeLabelDensity();
	path = [];
	for (var offset = labelDensity; offset < this.yAxisMaximum; offset += labelDensity)
		{
		path = path.concat([ 'M', this.leftMargin + 0.5, this.getY(offset), 'H', this.width + 0.5 ]);
		this.drawLabel(offset);
		}
	this.canvas.path(path.join()).attr({ stroke: '#ddd' });
	};

Graph.prototype.computeLabelDensity = function ()
	{
	var max = this.yAxisMaximum;
	var density = this.minUnit / 4;
	while (max / density > 8)
		{
		density *= 2;
		}
	return density;
	};

Graph.prototype.drawLabel = function (speed)
	{
	this.canvas
		.text(this.leftMargin / 2, this.getY(speed), Math.round(speed) + 'k')
		.attr({ font: '10px Helvetica, Arial', fill: "#aaa" });
	};

Graph.prototype.drawPoints = function (points, color)
	{
var     leftgutter = this.leftMargin,
        bottomgutter = 20,
        topgutter = 20;

	var count = points.length;
	var set = this.canvas.set();
	var prevX = this.leftMargin + 0.5;
	var prevY = this.getY(points[0]);
	var path = [ 'M', prevX, prevY, 'C', prevX, prevY ];
	var width = this.width - this.leftMargin;
	var x =  Math.round(width / (count - 1)) + this.leftMargin + 0.5;
	var y = this.getY(points[1]);

	for (var idx = 1; idx < count - 1; ++idx)
		{
		var nextX = Math.round((idx + 1) * width / (count - 1)) + this.leftMargin + 0.5;
		var nextY = this.getY(points[idx + 1]);
		var a = getAnchors(prevX, prevY, x, y, nextX, nextY);
		path = path.concat([ a.x1, a.y1, x, y, a.x2, a.y2 ]);
		prevX = x;
		prevY = y;
		x = nextX;
		y = nextY;
		set.push(this.canvas.rect(x - 1.5, y - 1.5, 3, 3).attr({ fill: color }));
		}
	path = path.concat([ x, y, x, y]);
	set.push(this.canvas.path(path.join()));
	return set.attr({ stroke: color, "stroke-width": 3, "stroke-linejoin": "round" });
	};

Graph.prototype.getY = function (y)
	{
	var height = this.height - this.topMargin;
	return Math.round(height - y * (height / this.yAxisMaximum)) + this.topMargin + 0.5;
	};
